var Objects = require("convenience").Objects;
var _ = require("lodash");
var SettableFuture = com.google.common.util.concurrent.SettableFuture;
var Executors = java.util.concurrent.Executors;

/*
 * Parse function parameter names.
 */
function params (func)
{
    var funcStr = func.toString();
    var parsedParams = funcStr.substring(
        funcStr.indexOf("(") + 1,
        funcStr.indexOf(")")
    ).split(",");

    var parameterNames = [];
    for (var i = 0; i < parsedParams.length; i++)
    {
        parsedParams[i] = parsedParams[i].trim();

        if (parsedParams[i] !== "")
        {
            parameterNames.push(parsedParams[i]);
        }
    }

    return parameterNames;
}

/*
 * Create an API handler for a javascript callback.
 */
function makeHandler (apiProxy, method, callback)
{
    var HttpHandler = org.webbitserver.HttpHandler;    
    var funcParams = params(callback);
    return new HttpHandler({
        handleHttpRequest: function(request, response, control) {
            var params = Objects.fromMap(method.getParameters(request));
            var funcArgs = [];

            for (var i = 0; i < funcParams.length; i++)
            {
                funcArgs.push(Objects.toJS(params[funcParams[i]], true));
            }
            
            var retval = callback.apply(apiProxy, funcArgs);
            
            if (retval instanceof Response)
            {
                retval.sendTo(request, response);
            }
            else
            {
                retval = JSON.stringify(retval);
                response.header("Content-Type","application/json").content(retval).end();                
            }
        }
    });
}

/*
 * Define an API method.
 */
function method (key, groupKey, group, apiProxy, descriptor)
{
    if (typeof descriptor === "function")
    {
        return method(key, groupKey, group, apiProxy, {
            callback: descriptor,
            verb: "post",
            name: key,
            endpoint: groupKey + "/" + key
        });
    }
    else
    {
        var endpoint = "endpoint" in descriptor ? descriptor.endpoint : groupKey + "/" + key;
        var description = "description" in descriptor ? descriptor.description : key;
        var verb = "verb" in descriptor ? descriptor.verb : "post";
        var auth = "auth" in descriptor ? descriptor.auth : ("require" in descriptor ? descriptor.require : undefined);
        var name = "name" in descriptor ? descriptor.name : key;
        var version = descriptor.version;
        
        var apiMethod = version ? group.method(name, endpoint, description, version)
                                : group.method(name, endpoint, description);
        
        apiMethod.verb(verb);
        
        if (auth)
        {
            apiMethod.requireAuth(auth);
        }
        
        if ("params" in descriptor)
        {
            for (var paramName in descriptor.params)
            {
                /* jshint loopfunc:true */
                (function(key, param) {
                    var type = "type" in param ? param.type : "any";
                    var description = "description" in param ? param.description : key;
                    if ('default' in param)
                    {
                        apiMethod.param(type, key, description, param['default']);                    
                    }
                    else
                    {
                        apiMethod.param(type, key, description);                    
                    }
                })(paramName, descriptor.params[paramName]);
                /* jshint loopfunc:false */
            }
        }
        
        if ("handler" in descriptor)
        {
            apiMethod.handler(descriptor.handler);
        }
        else if ("callback" in descriptor)
        {
            apiMethod.handler(makeHandler(apiProxy, apiMethod, descriptor.callback));
        }
        
        return apiMethod;
    }
}

/*
 * Return a wrapper for an API method that can be used to invoke as a javascript function.
 */
function wrapMethodAsFunction (apiProxy, method, descriptor)
{
    var callback = typeof descriptor === "function" ? descriptor : descriptor.callback;
    return function() { return callback(apiProxy, arguments); };
}

/*
 * An API response that can be returned from a callback.
 */
function Response (params)
{
    if (typeof params === "function")
    {
        this.sendTo = params;
    }
    else
    {
        this.sendTo = function (httpRequest, httpResponse)
        {
            var contentType = 'contentType' in params ? params.contentType : "text/plain";
            var headers = 'headers' in params ? params.headers : {};
            var status = 'status' in params ? params.status : 200;
            var content = params.content;

            httpResponse.status(status).header("Content-Type", contentType);

            for (var key in headers)
            {
                if (headers.hasOwnProperty(key))
                {
                    httpResponse.header(key, headers[key]);
                }
            }

            if (typeof content === "function")
            {
                content = content(httpRequest);
            }

            httpResponse.content(content).end();
        };            
    }
}

function PluginAPI (plugin)
{
    var apiBuilder = SettableFuture.create();
    var threadPool = Executors.newCachedThreadPool();

    if ('onEvent' in plugin)
    {
        plugin.onEvent(function(event) {
            if(event.getClass().toString().endsWith("BuildAPIEvent"))
            {
                apiBuilder.set(event.builder);
            }
        });
    }
    else if ('getClass' in plugin && plugin.getClass().toString().endsWith("APIBuilder"))
    {
        apiBuilder.set(plugin);
    }

    var API = function API(api)
    {
        var proxy = {};
        apiBuilder.addListener(function() {
            var builder = apiBuilder.get();
            for (var groupKey in api)
            {
                proxy[groupKey] = {};
                var groupDescriptor = api[groupKey];
                var title = 'title' in groupDescriptor ? groupDescriptor.title : groupKey;
                var description = 'description' in groupDescriptor ? groupDescriptor.description : "";
                var isPublic = 'isPublic' in groupDescriptor ? groupDescriptor.isPublic : true;
                var methods = 'methods' in groupDescriptor ? groupDescriptor.methods : {};
                var version = groupDescriptor.version;
                
                var group = builder.group(groupKey, title, description, version, isPublic);

                for (var methodKey in methods)
                {
                    var descriptor = methods[methodKey];
                    var apiMethod = method(
                        methodKey,
                        groupKey,
                        group,
                        proxy,
                        descriptor
                    );
                    proxy[groupKey][methodKey] = wrapMethodAsFunction(proxy, apiMethod, descriptor);
                }
            }            
        }, threadPool);
        return proxy;
    };

    API.Response = Response;
    return API;
}

module.exports = PluginAPI;