var facade = require("facade");
var Objects = require("convenience").Objects;
var _ = require("lodash");

var API = require("./apiHelper")(builder);

var File = java.io.File;
var FileWrapper = com.cattura.cattura_commons.vos.wrappers.FileWrapper;
var SystemInformationUtilities = com.cattura.agents.configuration_module.model.utilities.SystemInformationUtilities;
var Permissions = com.cattura.agents.configuration_module.model.properties.Permissions;
var Callback = com.cattura.share.utils.interfaces.Callback;

var PluginProxy = com.cattura.agents.configuration_module.model.proxies.PluginProxy;

var SETTINGS = facade.retrieveProxy("SettingsProxy");
var systemStatusProxy = facade.retrieveProxy("EnvironmentProxy");
var mediaProcessingProxy = facade.retrieveProxy("MediaProcessingProxy");
var captureStateProxy = facade.retrieveProxy("CaptureStateProxy");
var deviceProxy = facade.retrieveProxy("DeviceProxy");
var eventMonitoringProxy = facade.retrieveProxy("EventMonitoringProxy");
var userProxy = facade.retrieveProxy("UserProxy");

// Our json implementation, used to serialize Java objects
// that can't be passed to Rhino's json implementation
var json = new com.cattura.share.utils.json.parsing.JSON.Default();

var getUnitStatus = (function () {
    importPackage(java.util.concurrent);
    var status = {};
    var executor = Executors.newScheduledThreadPool(1);
    var updateFlags = {'storage': true, 'devices': true, 'capture': true, 'schedule': true, 'network': true, 'info': true};
    
    var eventHandlers = {
        CaptureStatusEvent: 'capture',
        SettingsUpdatedEvent: 'info',
        DevicesDetectedEvent: function () { 
            if (md5(api.status.getConnectedDevices()) != md5(status.devices.data))
            {
                updateFlags.devices = true;
            }
        },
        VideoDevicesDetectedEvent: function () {this.DevicesDetectedEvent();},
        AudioDevicesDetectedEvent: function () {this.DevicesDetectedEvent();}
    };
    
    function md5 (arg)
    {
        if (typeof arg == 'object')
        {
            arg = JSON.stringify(arg);
        }
        
        return com.google.common.hash.Hashing.md5().hashString(arg, com.google.common.base.Charsets.UTF_8).toString();
    }
    
    function startStatusWatcher ()
    {
        var networkWatcher = systemStatusProxy.networkStatusWatcher;
        var diskSpaceWatcher = systemStatusProxy.diskSpaceWatcher;
        
        diskSpaceWatcher.onUpdate(Objects.toConsumer(function() {
            updateFlags.storage = true;
        }));
        
        networkWatcher.onUpdate(Objects.toConsumer(function() {
            updateFlags.network = true;
        }));
        
        LOGGER.info("Starting status watcher thread");
        executor.scheduleAtFixedRate(updateStatus, 10, 10, TimeUnit.SECONDS);
        eventMonitoringProxy.addListener(function(event) {
            if (event.getClass().simpleName in eventHandlers)
            {
                var handler = eventHandlers[event.getClass().simpleName];
                if (typeof handler === 'function')
                {
                    handler.call(eventHandlers, event);
                }
                else if(handler.constructor === Array)
                {
                    for (var i = 0; i < handler; i++)
                    {
                        updateFlags[handler[i]] = true;
                    }
                }
                else
                {
                    updateFlags[handler] = true;
                }
            }
        });
    }
    
    function updateStatus ()
    {
        var now = java.lang.System.currentTimeMillis();
        
        if (updateFlags.storage)
        {
            status.storage = {lastUpdate: now, data: api.status.getDiskSpaceInfo()};
            updateFlags.storage = false;
        }

        if (updateFlags.network)
        {
            status.network = {lastUpdate: now, data: api.status.getNetworkInfo()};
            updateFlags.network = false;
        }

        if (updateFlags.devices)
        {
            status.devices = {lastUpdate: now, data: api.status.getConnectedDevices()};
            updateFlags.devices = false;
        }
        
        if (updateFlags.capture)
        {
            status.capture = {lastUpdate: now, data: api.status.getCaptureState()};
            updateFlags.capture = false;
        }
        
        if (updateFlags.info)
        {
            status.info = {lastUpdate: now, data: api.status.getUnitInfo()};
            updateFlags.info = false;
        }
        
        var schedule = api.status.getTodaysSchedule();
        updateFlags.schedule = updateFlags.schedule || md5(schedule) != md5(status.schedule.data);
        if (updateFlags.schedule)
        {
            status.schedule = {lastUpdate: now, data: schedule};
            updateFlags.schedule = false;
        }        
    }
    
    return function getUnitStatus () 
    {
        if (_(status).isEmpty())
        {
            java.lang.System.err.println("Creating status structure");
            startStatusWatcher();
            updateStatus();
        }
        
        return status;
    };
})();

LOGGER.info("Plugin loaded");
/*
 * The Hardware Unit Configuration API plugin provides APIs to manage the unit itself, i.e.
 * settings, users, plugins, etc.
 */
var api = API({
    /*
     * The Status API provides information about the current state of the unit,
     * i.e. what is going on, how much free space is left, etc.
     */
    status: {
        title: "Unit Status",
        description: "Provides methods to query the current state of the unit.",
        version: 1,
        methods: {
            general: {
                name: "getUnitStatus",
                description: "Obtain general information about the current state of the unit.",
                verb: "get",
                //require: "Administrator",
                endpoint: "status",
                params: {
                    since: {type: "long", description: "timestamp by which to compare status to.", "default": -1},
                },
                callback: function(since) {
                    var status = getUnitStatus();
                    var lastUpdate = -1;
                    
                    var returnStructure = {};
                    for (var key in status)
                    {
                        if (status[key].lastUpdate > since)
                        {
                            returnStructure[key] = status[key].data;
                            lastUpdate = Math.max(status[key].lastUpdate, lastUpdate);
                        }
                    }
                    
                    if (!_(returnStructure).isEmpty())
                    {
                        returnStructure.lastUpdate = lastUpdate;
                    }
                    
                    return returnStructure;
                }
            },
            
            unit: {
                name: "getUnitInfo",
                description: "Get basic information about the unit.",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "status/unit",
                callback: function () {
                    var updaterProxy = facade.retrieveProxy("UpdaterProxy");
                    var versionInfo = updaterProxy.getVersionInfo();

                    return Objects.toJS({
                        unitName: SETTINGS.get("unit-name"),
                        machineName: SystemInformationUtilities.getMachineName(),
                        uuid: SystemInformationUtilities.getSystemUUID(),
                        agentVersion: versionInfo.getCurrentVersion(),
                        serial: SETTINGS.getPrivateSettings().get("serial")
                    },true);
                }
            },

            network: {
                name: "getNetworkInfo",
                description: "Gets information about the network connectivity of the unit.",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "status/network",
                callback: function () {
                    var networkWatcher = systemStatusProxy.networkStatusWatcher;                    
                    return Objects.toJS({
                       internet: {
                           connected: networkWatcher.hasInternet(),
                           ip: networkWatcher.globalIP
                       },
                       local: {
                           ip: networkWatcher.localIP,
                           type: networkWatcher.networkType,
                           subnet: networkWatcher.subnet,
                           gateway: networkWatcher.gateway,
                           mac: networkWatcher.getMACAddress()
                       }
                    },true);
                }
            },

            space: {
                name: "getDiskSpaceInfo",
                description: "Gets the amount of free, usable, and total disk space.",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "status/space",
                callback: function () {
                    var diskSpaceWatcher = systemStatusProxy.diskSpaceWatcher;
                    return Objects.toJS({
                        total: diskSpaceWatcher.totalSpace,
                        usable: diskSpaceWatcher.usableSpace,
                        free: diskSpaceWatcher.freeSpace,
                        used: diskSpaceWatcher.usedSpace
                    },true);
                }
            },
            
            captureState: {
                name: "getCaptureState",
                description: "Returns the current capture's state (idle, capturing, processing, etc)",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "status/capture",
                callback: function () {
                    var captureState = captureStateProxy.captureState;
                    var stateStructure = {
                        state: captureState.state.name()
                    };
                                        
                    if (stateStructure.state != "IDLE")
                    {
                        stateStructure.currentCapture = {
                            title: captureState.captureRequest.title,
                            startTime: captureState.captureRequest.creationTime.time,
                            metadata: captureState.captureRequest.metadata,
                            tags: captureState.captureRequest.tags
                        };
                        
                        if (captureState.session)
                        {
                            stateStructure.currentCapture.paused = captureState.session.paused;
                            stateStructure.currentCapture.elapsed = captureState.session.elapsedTimeMillis;
                        }
                    }
                    
                    return Objects.toJS(stateStructure, true);
                }
            },
            
            connectedDevices: {
                name: "getConnectedDevices",
                description: "Returns a listing of audio and video devices connected to this unit.",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "status/devices",
                callback: function () {
                    var deviceStructure = {};
                    var audioDevices = Objects.toJS(deviceProxy.audioDevices);
                    var videoDevices = Objects.toJS(deviceProxy.videoDevices);
                    var key;
                    
                    for (key in audioDevices)
                    {
                        deviceStructure[key] = {
                            id: audioDevices[key].name,
                            name: audioDevices[key].prettyName,
                            type: "audio"
                        };
                    }
                    
                    for (key in videoDevices)
                    {
                        deviceStructure[key] = {
                            id: videoDevices[key].name,
                            name: videoDevices[key].prettyName,
                            type: "video",
                            snapshot: "/thumbnail/" + key,
                            liveStream: "/live/" + key,
                            width: videoDevices[key].width,
                            height: videoDevices[key].height,
                            fps: videoDevices[key].FPS
                        };
                    }
                    
                    return Objects.toJS(deviceStructure, true);
                }
            },
            
            todaysSchedule: {
                name: "getTodaysSchedule",
                description: "Returns all the scheduled captures for today.",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "status/schedule",
                callback: function () {
                    var schedulerProxy = facade.retrieveProxy("SchedulerProxy");            
                    var scheduleStructure = [];
                    var events = Objects.toJS(schedulerProxy.todaysEvents);

                    for (var i = 0; i < events.length; i++)
                    {
                        var event = events[i];
                        var task = Objects.toJS(schedulerProxy.getScheduler().getTaskSource().getTaskForEvent(event));
                        scheduleStructure.push({
                            id: event.ID,
                            group: event.groupID,
                            title: event.title,
                            start: event.startTime.time,
                            stop: event.stopTime.time,
                            templateID: task.templateID,
                            mediaPackageID: task.mediaPackageID
                        });
                    }
                    
                    return Objects.toJS(scheduleStructure, true);
                }
            }
        }
    },
    

   /*
    * The Settings API provides a means to programatically control the settings exposed through
    * the UI settings tab.
    */    
    settings: {
        title: "Settings",
        description: "Provides a programmatic way to manage the settings on the unit.",
        version: 1,
        methods: {
            getSetting: {
                name: "getSetting",
                description: "Returns a single setting for this unit.",
                verb: "get",
                require: Permissions.MANAGE_SETTINGS,            
                endpoint: "settings/{setting}",
                params: {
                    setting: {type: "string", description: "the name of the setting"},
                },
                callback: function (setting) {
                    var settings = Objects.toJS(SETTINGS.getPublicSettings(), true);
                    if (!(setting in settings))
                    {
                        return null;
                    }
                    
                    return settings[setting];
                }
            },
            setSetting: {
                name: "setSetting",
                description: "Update a single setting on this unit.",
                verb: "post",
                require: Permissions.MANAGE_SETTINGS,            
                endpoint: "settings/{setting}",
                params: {
                    setting: {type: "string", description: "the name of the setting"},
                    value: {description: "the value to set the setting to"}
                },                
                callback: function(setting,value) {
                    SETTINGS.set(setting,value);
                    return value;
                }
            },
            getAll: {
                name: "getSettings",
                description: "Returns all settings for the unit.",
                verb: "get",
                require: Permissions.MANAGE_SETTINGS,
                endpoint: "settings",
                callback: function() {
                    return Objects.toJS(SETTINGS.getPublicSettings(), true);
                }
            },
            setAll: {
                name: "setSettings",
                description: "Update multiple settings on the unit. Will not erase other settings",
                verb: "post",
                require: Permissions.MANAGE_SETTINGS,            
                endpoint: "settings",
                callback: function (settings) {
                    SETTINGS.set(JSON.parse(settings));
                }
            }
        }
    },    
    
    /*
     * The Users API provides programmatic control over the registered users
     * of the system.
     */
    users: {
        title: "Users",
        description: "Provides access to and methods to manage the users on the unit.",
        version: 1,
        methods: {
            get: {
                name: "getUsers",
                description: "List out all users registered on this box.",
                verb: "get",
                require: Permissions.MANAGE_USERS,
                endpoint: "users",
                callback: function() {
                    return _(Objects.toJS(userProxy.getUsers())).map(function(user) {
                        return {
                            name: Objects.toJS(user.username),
                            roles: Objects.toJS(user.roles, true),
                            email: Objects.toJS(user.emailAddress)
                        };
                    }).value();
                }
            }
        }
    },
    
    /*
     * The Maintenance API allows users to perform maintenance tasks on the
     * hardware unit remotely.
     */
    maintenance: {
        title: "Maintenance",
        description: "Maintenance tasks, such as rebooting.",
        version: 1,
        methods: {
            reboot: {
                name: "reboot",
                description: "Perform a reboot of this capture unit.",
                verb: "post",
                require: Permissions.MANAGE_SETTINGS,
                endpoint: "maintenance/reboot",
                callback: function() {
                    facade.sendNotification(facade.REBOOT);
                    return true;
                }
            },
            restart: {
                name: "restart",
                description: "Perform a restart of CaptureCast Pro.",
                version: "1.2",
                verb: "post",
                require: Permissions.MANAGE_SETTINGS,
                endpoint: "maintenance/restart",
                callback: function() {
                    facade.sendNotification(facade.RESTART);
                    return true;
                }
            },
            restartMediaProcessingModule: {
                name: "restartMediaProcessingModule",
                description: "Restarts only the media processing components",
                version: "1.2",
                verb: "post",
                require: Permissions.MANAGE_SETTINGS,
                endpoint: "maintenance/restartMediaProcessingModule",
                callback: function() {
                    facade.sendNotification(facade.RESTART_MEDIA_PROCESSING_MODULE);
                    return true;
                }
            }
        }
    },
    
    /*
     * The Plugins API provides information about, and eventually will allow
     * the administrator of the unit to install and update plugins.
     */
    plugins: {
        title: "Plugins",
        description: "Provides information about, and methods to manage, plugins installed on the capture unit.",
        version: 1,
        methods: {
            get: {
                name: "getPlugins",
                description: "Returns a list of all plugins installed on the agent.",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "plugins",
                callback: function() {
                    var mediaProcessingModule = facade.retrieveProxy("MediaProcessingProxy").getCaptureModule();
                    var pluginProxy = facade.retrieveProxy(PluginProxy.NAME);
                    var plugins = {};

                    /*
                     * MPA module plugins
                     */
                    if (mediaProcessingModule && mediaProcessingModule.hasFoundPlugins())
                    {
                        _(Objects.toJS(mediaProcessingModule.getPlugins())).each(function(plugin) {
                            var services = {};

                            _(Objects.toJS(plugin.services)).each(function(service) {
                                var parameters = {};

                                _(Objects.toJS(service.parameters)).map(function(parameter) {
                                    parameters[parameter.name] = {
                                        name: parameter.name,
                                        description: parameter.description,
                                        type: parameter.type
                                    };
                                }).value();

                                services[service.name] = {
                                    name: service.name,
                                    description: service.description,
                                    parameters: parameters
                                };
                            });

                            plugins[plugin.ID] = {
                                id: plugin.ID,
                                type: plugin.type.toString(),                        
                                title: plugin.title,
                                module: "media-processing-agent",
                                description: plugin.description,
                                copyright: plugin.copyright,
                                options: JSON.parse(plugin.optionsAsString.get().content),
                                services: services,
                                inputInterests: plugin.inputInterests,
                                outputInterests: plugin.outputInterests
                            };
                        });
                    }

                    /*
                     * Configuration Module plugins
                     */
                    _(Objects.toJS(pluginProxy.getPluginInfos())).each(function(plugin) {
                        var services = {};

                        _(Objects.toJS(plugin.services)).each(function(service) {
                            var parameters = {};

                            _(Objects.toJS(service.parameters)).map(function(parameter) {
                                parameters[parameter.name] = {
                                    name: parameter.name,
                                    description: parameter.description,
                                    type: parameter.parameterType.name
                                };
                            }).value();

                            services[service.name] = {
                                name: service.name,
                                description: service.description,
                                parameters: parameters
                            };
                        });
                        
                        plugins[plugin.pluginID] = {
                            id: plugin.pluginID,
                            type: "configuration", // might need to be able to specify this
                            title: plugin.title,
                            module: "configuration",
                            description: plugin.description,
                            copyright: plugin.copyright,
                            services: services,
                            options: {} // Eventually we'll define these for CM plugins
                        };
                    });

                    return Objects.toJS(plugins, true);
                }
            },

            getSettingsForPlugin: {
                name: "getSettingsForPlugin",
                description: "Retrieve user settings for a plugin.",
                verb: "get",
                require: Permissions.VIEW,
                endpoint: "plugins/{plugin}/settings",
                params: {
                    plugin: {type: "string", description: "the ID of the plugin"}
                },
                callback: function(plugin) {
                    var pluginSettings = SETTINGS.getUserSettingsForPluginId(plugin).orNull();
                    
                    if (pluginSettings)
                    {
                        return Objects.toJS(pluginSettings.toMap(), true);
                    }
                    
                    return {};                    
                }
            },
            
            setSettingsForPlugin: {
                name: "setSettingsForPlugin",
                description: "Retrieve user settings for a plugin.",
                verb: "post",
                require: Permissions.MANAGE_SETTINGS,
                endpoint: "plugins/{plugin}/settings",
                params: {
                    plugin: {type: "string", description: "the ID of the plugin to invoke the service on"},
                    settings: {type: "json", description: "key/value mapping of settings to set on the plugin"}                    
                },
                callback: function(plugin, settings) {
                    var pluginSettings = SETTINGS.getUserSettingsForPluginId(plugin).orNull();
                    settings = JSON.parse(settings);
                    
                    if (pluginSettings)
                    {
                        _(settings).each(function (value, key) {
                            pluginSettings.set(key, value);
                        });
                        return Objects.toJS(pluginSettings.toMap(), true);
                    }
                    
                    return {};                    
                }
            },

            callService: {
                name: "callPluginService",
                description: "Invoke a service provided by a plugin",
                verb: "post",
                require: Permissions.VIEW,
                endpoint: "plugins/{plugin}/{service}",
                params: {
                    plugin: {type: "string", description: "the ID of the plugin to invoke the service on"},
                    service: {type: "string", description: "the name of the service to call"},
                    args: {type: "json", description: "arguments for the service, as a JSON string", "default": "{}"}
                },
                callback: function(plugin, service, args) {
                    var pluginProxy = facade.retrieveProxy(PluginProxy.NAME);
                    var mediaProcessingModule = facade.retrieveProxy("MediaProcessingProxy").getCaptureModule();
                    args = JSON.parse(args);
                    var pluginName = plugin;
                    
                    if (mediaProcessingModule && mediaProcessingModule.hasFoundPlugins())
                    {
                        plugin = mediaProcessingModule.getPlugin(pluginName);
                        if (plugin)
                        {
                            var result = plugin.getService(service)['invoke(java.util.Map)'](args);
                            return JSON.parse(json.stringify(result).getOrRethrow());
                        }
                    }
                    
                    plugin = pluginProxy.getPluginInfo(pluginName);
                    if (plugin.isPresent())
                    {
                        var services = Objects.toJS(plugin.get().getServicesByName(service)),
                            preferredService = null;
                        
                        for (var i = 0; i < services.length; i++)
                        {
                            if (!services[i]['canInvoke(java.util.Map)'](args)) continue;
                            if (preferredService === null) preferredService = services[i];
                            if (services[i].getParameters().size() === _(args).size())
                            {
                                preferredService = services[i];
                                break;
                            }
                        }
                        
                        if (preferredService !== null)
                        {
                            return JSON.parse(json.stringify(preferredService['invoke(java.util.Map)'](args)).getOrRethrow()); //Objects.toJS(result,true);                            
                        }
                    }
                    
                    return null;
                }
            }
        }
    },
   
    /*
     * The Devices API gives programmatic access to information about
     * devices connected to the unit, and also allows redetection.
     */
    devices: {
        title: "Devices",
        description: "Provides information about, and methods to work with, audio and video devices.",
        version: 1,
        methods: {
            getAudioDevices: {
                name: "getAudioDevices",
                description: "Returns a listing of known audio devices.",
                verb: "get",
                endpoint: "devices/audio",
                callback: function() {
                    return JSON.parse(json.stringify(deviceProxy.getAudioDevices()).getOrRethrow());
                }
            },
            getVideoDevices: {
                name: "getVideoDevices",
                description: "Returns a listing of known video devices.",
                verb: "get",
                endpoint: "devices/video",
                callback: function() {
                    return JSON.parse(json.stringify(deviceProxy.getVideoDevices()).getOrRethrow());
                }
            },
            redetect: {
                name: "redetect",
                description: "Redetect all devices on the unit.",
                verb: "post",
                endpoint: "devices/redetect",
                require: Permissions.MANAGE_DEVICES,
                callback: function() {
                    deviceProxy.redetect();
                    return true;
                }
            },
            redetectProgress: {
                name: "redetectProgress",
                description: "Redetect all devices on the unit.",
                verb: "get",
                endpoint: "devices/redetect/progress",
                callback: function() {
                    return deviceProxy.getDetectionProgress();
                }
            }
        }
    }
});